using Microsoft.Data.SqlClient;
using Microsoft.SqlServer.Management.Common;
using Microsoft.SqlServer.Management.Smo;
using System.Data;
using Excel = Microsoft.Office.Interop.Excel;

public partial class SearchInDatabase : Form
{
    string _database = string.Empty;
    ServerConnection? _serverConnection;
    Server? _server;
    Database? _db;
    SqlConnection? _conn;
    string _connstr = string.Empty;

    const int MaxExcelRows = 1048000; // We consider the maximum number of rows less than 1,048,576

    string _databaseName = string.Empty;
    string _tableName = string.Empty;

    string _stringToSearch = string.Empty;

    // This is the search result
    // The key is table name, the list contains columns where the string was found
    Dictionary<string, List<string>> _foundTables = new Dictionary<string, List<string>>();

    public SearchInDatabase()
    {
        InitializeComponent();
    }

    private void exit_Click(object sender, EventArgs e)
    {
        Application.Exit();
    }

    private void connect_Click(object sender, EventArgs e)
    {
        if(sqlServerBox.Text.Length == 0)
            return;

        _connstr = "Data Source=" + sqlServerBox.Text + ";Integrated Security=true;Column Encryption Setting=enabled;MultipleActiveResultSets=True;TrustServerCertificate=True";
        try
        {
            _conn = new SqlConnection(_connstr);
            _serverConnection = new ServerConnection(_conn);
            _serverConnection.Connect();
            _server = new Server(_serverConnection);
            _server.ConnectionContext.AutoDisconnectMode = AutoDisconnectMode.NoAutoDisconnect;

            DatabaseCollection dbs = _server.Databases;

            foreach(Database db in dbs)
            {
                if(!db.IsSystemObject) // Add only user databases
                    databaseList.Items.Add(db.Name);
            }

            label2.Visible = true;
            searchText.Visible = true;
            caseSensitive.Visible = true;
            label3.Visible = true;
            databaseList.Visible = true;
        }
        catch(SmoException ex)
        {
            MessageBox.Show("Cannot connect to SQL server " + sqlServerBox.Text + "\n" + ex.Message);
        }
        catch(Exception ex)
        {
            MessageBox.Show("Cannot connect to SQL server " + sqlServerBox.Text + "\n" + ex.Message);
        }
    }

    private void databaseList_SelectedIndexChanged(object sender, EventArgs e)
    {
        if(databaseList.SelectedItem == null) // This should NEVER happen, because we selected an item previously
            return;
        try
        {
            // We should never get an error here!
            _databaseName = databaseList.SelectedItem.ToString();
            _db = _server.Databases[_databaseName];
            TableCollection tables = _db.Tables;

            foreach(Table tbl in tables)
            {
                if(!tbl.IsSystemObject)
                    tableList.Items.Add(tbl.Name);
            }

            if(tableList.Items.Count == 0)
            {
                MessageBox.Show("There are no user tables in  database [" + _databaseName + "]\n");
                return;
            }

            label4.Visible = true;
            tableList.Visible = true;
        }
        catch(SmoException ex)
        {
            MessageBox.Show("Cannot get tables list from database [" + _databaseName + "]\n" + ex.Message);
        }
        catch(Exception ex)
        {
            MessageBox.Show("Cannot get tables list from database [" + _databaseName + "]\n" + ex.Message);
        }
    }

    private void tableList_SelectedIndexChanged(object sender, EventArgs e)
    {
        try
        {
            _tableName = tableList.SelectedItem.ToString();
            search.Visible = true;
        }
        catch(SmoException ex)
        {
            MessageBox.Show("Cannot get table info for table [" + _tableName + "]\n" + ex.Message);
        }
        catch(Exception ex)
        {
            MessageBox.Show("Cannot get table info for table [" + _tableName + "]\n" + ex.Message);
        }
    }

    private void search_Click(object sender, EventArgs e)
    {
        if(tableList.SelectedItem == null) // This should NEVER happen!
        {
            search.Visible = false;
            return;
        }
        if(searchText.Text.Length == 0)
        {
            MessageBox.Show("You have to enter the string you search");
            searchText.Focus();
            return;
        }
        _stringToSearch = searchText.Text;
        try
        {
            SearchInTable(_db.Tables[_tableName]);
        }
        catch(SmoException ex)
        {
            MessageBox.Show("Cannot search in table [" + _tableName + "]\n" + ex.Message);
        }
        catch(Exception ex)
        {
            MessageBox.Show("Cannot search in table [" + _tableName + "]\n" + ex.Message);
        }
    }

    private void SearchInTable(Table tbl)
    {
        string sqlcmd = "select * from [" + tbl.Name + "]";
        using DataSet result = _db.ExecuteWithResults(sqlcmd);
        ColumnCollection columns = tbl.Columns;
        Dictionary<string, int> columnsFound = new Dictionary<string, int>();
        int lastColumnIndex = -1; // last index of the column in the dictionary

        string fileName = Environment.GetFolderPath(Environment.SpecialFolder.Desktop) + "\\" + _databaseName + "-" + tbl.Name + ".xlsx";

        Excel.Application xlsApp;
        Excel.Workbook xlsWorkbook;
        Excel.Worksheet xlsWorksheet;
        object misValue = System.Reflection.Missing.Value;

        /*
         * We will export found column in an Excel file in the following way:
         * row - contains the row where we found the string
         * columns - are columns which contain the table field where the string has been found
         * 
         * In order to maintain an easy manipulation we will have a list with column namess 
         * we already found with the information
         * If a column name is new we add it to the list and fill in the corresponding cell
         * in the Excell sheet
         * If a column has ben found (it is in the list) but for the current row it doesn't
         * contain the string we will let it empty in the Excel file
         */
        try
        {
            bool foundstring = false;
            xlsApp = new Excel.Application();
            xlsWorkbook = xlsApp.Workbooks.Add(misValue);
            xlsWorksheet = (Excel.Worksheet)xlsWorkbook.Sheets[1];

            // Create the header for Excel file
            xlsWorksheet.Cells[1, 1] = "Search string in a database table report";
            Excel.Range range = xlsWorksheet.get_Range("A1", "E1");
            range.Merge(1);
            range.Borders.Color = Color.Black.ToArgb();
            range.Interior.Color = Color.Yellow.ToArgb();

            xlsWorksheet.Cells[2, 1] = "String to search";
            xlsWorksheet.Cells[2, 2] = _stringToSearch;

            xlsWorksheet.Cells[3, 1] = "Search in table";
            xlsWorksheet.Cells[3, 2] = tbl.Name;

            // We write column names in the 3rd row of the Excel file
            // Data will be written from row 4 on
            int i = 6;
            foreach(DataRow row in result.Tables[0].Rows)
            {
                bool haverows = false;
                foreach(Column col in columns)
                {
                    // Ignore columns which are not string
                    if(!col.DataType.IsStringType)
                        continue;

                    // Ignore columns which are shorter than what we search
                    if(_stringToSearch.Length > col.DataType.MaximumLength)
                        continue;

                    // If the column in this row is null, ignore it
                    if(row[col.Name] == DBNull.Value)
                        continue;

                    string cell = (string)row[col.Name];
                    if(cell != null)
                    {
                        if(cell.Contains(_stringToSearch, (caseSensitive.Checked) ?
                            StringComparison.InvariantCulture : StringComparison.InvariantCultureIgnoreCase))
                        {
                            // found the string
                            haverows = true;
                            foundstring = true;
                            int colindex;
                            // see if the column we already found
                            if(!columnsFound.ContainsKey(col.Name))
                            {
                                // Add the new column to the list and to the Excel file
                                columnsFound.Add(col.Name, ++lastColumnIndex);
                                xlsWorksheet.Cells[5, lastColumnIndex + 1] = col.Name;
                                colindex = lastColumnIndex + 1;
                            }
                            else
                            {
                                // Get the Excel column index
                                colindex = columnsFound[col.Name] + 1;
                            }
                            xlsWorksheet.Cells[i, colindex] = cell.Length <= 255 ? cell : "<Table column too is long>";
                        }
                    }
                }
                if(haverows)
                    ++i; // go to the next row in the Excel file
                if(i > MaxExcelRows) // We have too many rows in the table
                {
                    break;
                }
            }
            if(foundstring)
            {
                // Remove the old excel report file
                try
                {
                    FileInfo oldFile = new FileInfo(fileName);
                    if(oldFile.Exists)
                    {
                        File.SetAttributes(oldFile.FullName, FileAttributes.Normal);
                        oldFile.Delete();
                    }
                }
                catch(Exception ex)
                {
                    MessageBox.Show("Error removing old Excel report: " + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Stop);
                    return;
                }

                // we found the string, try to save the Excel file
                xlsWorkbook.SaveAs(fileName, Excel.XlFileFormat.xlWorkbookDefault, misValue, misValue,
                    misValue, misValue,
                    Excel.XlSaveAsAccessMode.xlExclusive, Excel.XlSaveConflictResolution.xlLocalSessionChanges,
                    misValue, misValue, misValue, misValue);
                xlsWorkbook.Close(true, misValue, misValue);
                MessageBox.Show("String [" + _stringToSearch + "]\nwas found in one or more columns in the table [" +
                    tbl.Name + "]\nCheck the [" + fileName + "] for results", "String was found",
                    MessageBoxButtons.OK, MessageBoxIcon.Information);
            }
            else
            {
                xlsWorkbook.Close(false, misValue, misValue);
                // We didn't find the string in the selected table
                // In this case we will not save the Excel file
                MessageBox.Show("String [" + _stringToSearch + "]\nwas not found in the table [" +
                    tbl.Name + "]", "String was not found",
                    MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            xlsApp.Quit();

            ReleaseObject(xlsWorksheet);
            ReleaseObject(xlsWorkbook);
            ReleaseObject(xlsApp);
        }
        catch(Exception ex)
        {
            MessageBox.Show("Error creating Excel report: " + ex.Message, "Error",
                MessageBoxButtons.OK, MessageBoxIcon.Stop);
        }
    }

    static private void ReleaseObject(object obj)
    {
        try
        {
            System.Runtime.InteropServices.Marshal.ReleaseComObject(obj);
            obj = null;
        }
        catch(Exception ex)
        {
            obj = null;
            MessageBox.Show("Exception occured while releasing object " + ex.ToString());
        }
        finally
        {
            GC.Collect();
        }
    }

}
